/*
 *  linux/kernel/system_call.s
 *
 *  (C) 1991  Linus Torvalds
 */

/*
 *  system_call.s  contains the system-call low-level handling routines.
 * This also contains the timer-interrupt handler, as some of the code is
 * the same. The hd- and flopppy-interrupts are also here.
 *
 * NOTE: This code handles signal-recognition, which happens every time
 * after a timer-interrupt and after each system call. Ordinary interrupts
 * don't handle signal-recognition, as that would clutter them up totally
 * unnecessarily.
 *
 * Stack layout in 'ret_from_system_call':
 *
 *	 0(%esp) - %eax
 *	 4(%esp) - %ebx
 *	 8(%esp) - %ecx
 *	 C(%esp) - %edx
 *	10(%esp) - %fs
 *	14(%esp) - %es
 *	18(%esp) - %ds
 *	1C(%esp) - %eip
 *	20(%esp) - %cs
 *	24(%esp) - %eflags
 *	28(%esp) - %oldesp
 *	2C(%esp) - %oldss
 */
/*
 * system_call.s ļϵͳ(system-call)ײ㴦ӳЩȽƣ
 * ͬʱҲʱжϴ(timer-interrupt)Ӳ̵̺жϴҲ
 *
 * ע⣺δ봦ź(signal)ʶÿʱжϺϵͳ֮󶼻ʶһ
 * жźŲźʶΪϵͳɻҡ
 *
 * ϵͳ÷أ'ret_from_system_call'ʱջݡ
 */

SIG_CHLD	= 17					# SIG_CHLD źţӽֹͣ

EAX			= 0x00					# ջиĴƫλá
EBX			= 0x04
ECX			= 0x08
EDX			= 0x0C
FS			= 0x10
ES			= 0x14
DS			= 0x18
EIP			= 0x1C
CS			= 0x20
EFLAGS		= 0x24
OLDESP		= 0x28					# Ȩ仯ʱջлûջָ뱻ں̬ջ
OLDSS		= 0x2C

# Щṹ(task_struct)бƫֵμinclude/linux/sched.hʼ
state	= 0			# these are offsets into the task-struct.	# ״̬
counter	= 4			# ʱ(ݼ)δʱƬ
priority = 8		# ʼʱcounter=priorityԽʱԽ
signal	= 12		# źλͼÿλһźţźֵ=λƫֵ+1
sigaction = 16		# MUST be 16 (=len of sigaction)  #sigaction ṹȱ16 ֽڡ
					# źִԽṹƫֵӦźŽҪִеĲͱ־Ϣ
blocked = (33*16)	# źλͼƫ

# ¶sigaction ṹеƫμinclude/signal.h
# offsets within sigaction
sa_handler = 0		# źŴ̵ľ
sa_mask = 4			# ź
sa_flags = 8		# źż
sa_restorer = 12	# ػִָеĵַλáμkernel/signal.c

nr_system_calls = 87	# Linux 0.11 ںеϵͳ

/*
 * Ok, I get parallel printer interrupts while using the floppy for some
 * strange reason. Urgel. Now I just ignore them.
 */
/*
 * ˣʹʱյ˲дӡжϣ֡ǣڲ
 */
# ڵ㡣
.globl _system_call,_sys_fork,_timer_interrupt,_sys_execve
.globl _hd_interrupt,_floppy_interrupt,_parallel_interrupt
.globl _device_not_available, _coprocessor_error

# ϵͳúš
.align 2						# ڴ4 ֽڶ롣
bad_sys_call:
	movl $-1,%eax				# eax -1˳жϡ
	iret
.align 2

# ִеȳڡȳschedule (kernel/sched.c)
reschedule:
	pushl $ret_from_sys_call	# ret_from_sys_call ĵַջ
	jmp _schedule
	
# int 0x80 --linux ϵͳڵ(жint 0x80eax ǵú)
.align 2
_system_call:
	cmpl $nr_system_calls-1,%eax	# úΧĻeax -1 ˳
	ja bad_sys_call
	push %ds						# ԭμĴֵ
	push %es
	push %fs
	pushl %edx			# ebx,ecx,edx зϵͳӦC Ժĵò
	pushl %ecx			# push %ebx,%ecx,%edx as parameters
	pushl %ebx			# to the system call
	movl $0x10,%edx		# set up ds,es to kernel space
	mov %dx,%ds			# ds,es ָںݶ(ȫݶ)
	mov %dx,%es
	# עLinux0.11ں˸ĴڴصġǵĶλַͶ޳ͬ μfork.ccopy_mem()
	movl $0x17,%edx		# fs points to local data space
	mov %dx,%fs			# fs ֲָݶ(ֲݶ)
# ĺǣõַ = _sys_call_table + %eax * 4
# ӦC еsys_call_table include/linux/sys.h Ужһ87 
# ϵͳC ĵַ
	call *_sys_call_table(,%eax,4)
	pushl %eax			# ѷֵջ
	movl _current,%eax	# ȡǰ񣨽̣ݽṹַ->eax
# в鿴ǰ״̬ھ״̬(state 0)ȥִеȳ
# ھ״̬counterֵ0Ҳȥִеȳ
	cmpl $0,state(%eax)			# state
	jne reschedule
	cmpl $0,counter(%eax)		# counter
	je reschedule
	
# δִдϵͳC غ󣬶źʶ
ret_from_sys_call:
	# бǰǷǳʼtask0򲻱ضźĴֱӷء
	# _task ӦC еtask[]飬ֱtask ൱task[0]
	movl _current,%eax			# task[0] cannot have signals
	cmpl _task,%eax
	je 3f
	# ͨԭóѡļжϵóǷǳûǳûֱ
	# ˳жϣźĴȽѡǷΪͨûεѡ0x000f
	# (RPL=3ֲ1 ())ת˳жϳ
	cmpw $0x0f,CS(%esp)			# was old code segment supervisor ?
	jne 3f
	# ԭջѡΪ0x17ҲԭջûݶУҲ˳
	cmpw $0x17,OLDSS(%esp)		# was stack segment = 0x17 ?
	jne 3f
# δ;ȡǰṹеźλͼ(32 λÿλ1 ź)
# ȻṹеźΣ룬źλȡֵСźֵٰ
# ԭźλͼиźŶӦλλ0󽫸źֵΪ֮һdo_signal()
# do_signal()ڣkernel/signal.cУ13 ջϢ
	movl signal(%eax),%ebx		# ȡźλͼ->ebxÿ1 λ1 źţ32 źš
	movl blocked(%eax),%ecx		# ȡΣźλͼ->ecx
	notl %ecx					# ÿλȡ
	andl %ebx,%ecx				# ɵźλͼ
	bsfl %ecx,%ecx				# ӵλλ0ʼɨλͼǷ1 λ
								# Уecx λƫֵڼλ0-31
	je 3f						# ûźǰת˳
	btrl %ecx,%ebx				# λźţebx ԭsignal λͼ
	movl %ebx,signal(%eax)		# ±signal λͼϢ->current->signal
	incl %ecx					# źŵΪ1 ʼ(1-32)
	pushl %ecx					# źֵջΪdo_signal Ĳ֮һ
	call _do_signal				# C źŴ(kernel/signal.c)
	popl %eax					# źֵ
3:	popl %eax
	popl %ebx
	popl %ecx
	popl %edx
	pop %fs
	pop %es
	pop %ds
	iret

#### int16 -- δ봦ЭĳźšתִC math_error()
# (kernel/math/math_emulate.c)غתret_from_sys_call ִС
.align 2
_coprocessor_error:
	push %ds
	push %es
	push %fs
	pushl %edx
	pushl %ecx
	pushl %ebx
	pushl %eax
	movl $0x10,%eax				# ds,es ΪָںݶΡ
	mov %ax,%ds
	mov %ax,%es
	movl $0x17,%eax				# fs ΪֲָݶΣݶΣ
	mov %ax,%fs
	pushl $ret_from_sys_call	# ÷صĵַջ
	jmp _math_error				# ִC math_error()(kernel/math/math_emulate.c)

#### int7 -- 豸ڻЭ(Coprocessor not available)
# ƼĴCR0 EM ־λCPU ִһESC תָʱͻжϣ
# лжϴģESC תָ
# CR0 TS ־CPU ִתʱõġTS ȷʲôʱЭеݣģ
# CPU ִеƥˡCPU һתָʱTS λˣͻжϡ
# ʱӦûָЭִ״̬μ(kernel/sched.c)е˵
# жתƵret_from_sys_call ִȥⲢźţ
.align 2
_device_not_available:
	push %ds
	push %es
	push %fs
	pushl %edx
	pushl %ecx
	pushl %ebx
	pushl %eax
	movl $0x10,%eax				# ds,es ΪָںݶΡ
	mov %ax,%ds
	mov %ax,%es
	movl $0x17,%eax				# fs ΪֲָݶΣݶΣ
	mov %ax,%fs
	pushl $ret_from_sys_call	# תõķصַջ
	clts						# clear TS so that we can use math
	movl %cr0,%eax
	testl $0x4,%eax				# EM (math emulation bit)
	# EM жϣָЭ״̬
	je _math_state_restore		# ִC math_state_restore()(kernel/sched.c)
	pushl %ebp
	pushl %esi
	pushl %edi
	call _math_emulate			# C math_emulate(kernel/math/math_emulate.c)
	popl %edi
	popl %esi
	popl %ebp
	ret							# ret תret_from_sys_call

#### int32 -- (int 0x20) ʱжϴжƵʱΪ100Hz(include/linux/sched.h)
# ʱоƬ8253/8254 (kernel/sched.c)ʼġjiffies ÿ10 1
# δ뽫jiffies 1ͽжָ8259 ȻõǰȨΪ
# C do_timer(long CPL)÷ʱתȥⲢźš
.align 2
_timer_interrupt:
	push %ds		# save ds,es and put kernel data space
	push %es		# into them. %fs is used by _system_call
	push %fs
	pushl %edx		# we save %eax,%ecx,%edx as gcc doesn't
	pushl %ecx		# save those across function calls. %ebx
	pushl %ebx		# is saved as we use that in ret_sys_call
	pushl %eax
	movl $0x10,%eax		# ds,es ΪָںݶΡ
	mov %ax,%ds
	mov %ax,%es
	movl $0x17,%eax		# fs ΪֲָݶΣݶΣ
	mov %ax,%fs
	incl _jiffies
	# ڳʼжϿоƬʱûвԶEOIҪָӲжϡ
	movb $0x20,%al		# EOI to interrupt controller #1
	outb %al,$0x20		# OCW2 0x20 ˿ڡ
	# 3 ѡȡǰȨ(0 3)ѹջΪdo_timer Ĳ
	movl CS(%esp),%eax
	andl $3,%eax		# %eax is CPL (0 or 3, 0=supervisor)
	pushl %eax
	# do_timer(CPL)ִлʱȹkernel/shched.c,ʵ֡
	call _do_timer		# 'do_timer(long CPL)' does everything from
	addl $4,%esp		# task switching to accounting ...
	jmp ret_from_sys_call

#### sys_execve()ϵͳáȡжϵóĴָΪC do_execve()
# do_execve()(fs/exec.c)
.align 2
_sys_execve:
	lea EIP(%esp),%eax
	pushl %eax
	call _do_execve
	addl $4,%esp			# ʱѹջEIP ֵ
	ret

#### sys_fork()ãڴӽ̣system_call 2ԭinclude/linux/sys.h С
# ȵC find_empty_process()ȡһ̺pidظ˵Ŀǰ
# Ȼcopy_process()ƽ̡
.align 2
_sys_fork:
	call _find_empty_process			# find_empty_process()(kernel/fork.c)
	testl %eax,%eax
	js 1f
	push %gs
	pushl %esi
	pushl %edi
	pushl %ebp
	pushl %eax
	call _copy_process					# C copy_process()(kernel/fork.c)
	addl $20,%esp						# ѹջݡ
1:	ret

#### int 46 -- (int 0x2E) ӲжϴӦӲжIRQ14
# Ӳ̲ɻͻᷢжźš(μkernel/blk_drv/hd.c)
# 8259A жϿƴоƬͽӲжָ(EOI)Ȼȡdo_hd еĺָedx
# ĴУdo_hd ΪNULLжedx ָǷΪաΪգedx ֵָ
# unexpected_hd_interrupt()ʾϢ8259A оƬEOI ָedx 
# ָָĺ: read_intr()write_intr()unexpected_hd_interrupt()
_hd_interrupt:
	pushl %eax
	pushl %ecx
	pushl %edx
	push %ds
	push %es
	push %fs
	movl $0x10,%eax						# ds,es ΪںݶΡ
	mov %ax,%ds
	mov %ax,%es
	movl $0x17,%eax						# fs ΪóľֲݶΡ
	mov %ax,%fs
	# ڳʼжϿоƬʱûвԶEOIҪָӲжϡ
	movb $0x20,%al
	outb %al,$0xA0		# EOI to interrupt controller #1 #ʹ8259A.
	jmp 1f				# give port chance to breathe
1:	jmp 1f				# ʱá
1:	xorl %edx,%edx
	xchgl _do_hd,%edx	# do_hd Ϊһָ룬ֵread_intr()
						# write_intr()ַ(kernel/blk_drv/hd.c)
						# ŵedx Ĵͽdo_hd ָΪNULL
	testl %edx,%edx		# ԺָǷΪNull
	jne 1f				# գʹָָC unexpected_hd_interrupt()
	movl $_unexpected_hd_interrupt,%edx
1:	outb %al,$0x20		# 8259A жϿEOI ָӲжϣ
	call *%edx			# "interesting" way of handling intr.
	pop %fs				# Ͼdo_hd ָC 
	pop %es
	pop %ds
	popl %edx
	popl %ecx
	popl %eax
	iret

#### int38 -- (int 0x26) жϴӦӲжIRQ6
# 䴦Ӳ̵Ĵһ(kernel/blk_drv/floppy.c)
# 8259A жϿоƬEOI ָȻȡdo_floppy еĺָeax
# ĴУdo_floppy ΪNULLжeax ָǷΪաΪգeax ֵָ
# unexpected_floppy_interrupt ()ʾϢeax ָĺ: rw_interrupt,
# seek_interrupt,recal_interrupt,reset_interrupt unexpected_floppy_interrupt
_floppy_interrupt:
	pushl %eax
	pushl %ecx
	pushl %edx
	push %ds
	push %es
	push %fs
	movl $0x10,%eax			# ds,es ΪںݶΡ
	mov %ax,%ds
	mov %ax,%es
	movl $0x17,%eax			# fs ΪóľֲݶΡ
	mov %ax,%fs
	movb $0x20,%al			# 8259A жϿEOI ָӲжϣ
	outb %al,$0x20			# EOI to interrupt controller #1
	xorl %eax,%eax
	xchgl _do_floppy,%eax	# do_floppy Ϊһָ룬ֵʵʴC 
							# ŵeax Ĵͽdo_floppy ָÿա
	testl %eax,%eax
	jne 1f
	movl $_unexpected_floppy_interrupt,%eax
1:	call *%eax				# "interesting" way of handling intr.
	pop %fs
	pop %es
	pop %ds
	popl %edx
	popl %ecx
	popl %eax
	iret

#### int 39 -- (int 0x27) пжϴ򣬶ӦӲжźIRQ7
# 汾ں˻δʵֻ֡ǷEOI ָ
_parallel_interrupt:
	pushl %eax
	movb $0x20,%al
	outb %al,$0x20
	popl %eax
	iret
